from enum import Enum

from meross_iot.cloud.abilities import *
from meross_iot.cloud.device import AbstractMerossDevice
from meross_iot.cloud.devices.light_bulbs import MODE_RGB, to_rgb, MODE_LUMINANCE, MODE_TEMPERATURE
from meross_iot.logger import HUMIDIFIER_LOGGER as l
from meross_iot.meross_event import HumidifierSpryEvent, HumidifierLightEvent
from meross_iot.utilities.conversion import int_to_rgb


class SprayMode(Enum):
    OFF = 0
    CONTINUOUS = 1
    INTERMITTENT = 2


class GenericHumidifier(AbstractMerossDevice):
    _raw_state = None

    def __init__(self, cloud_client, device_uuid, **kwords):
        super(GenericHumidifier, self).__init__(cloud_client, device_uuid, **kwords)

    def _handle_push_notification(self, namespace, payload, from_myself=False) -> bool:
        if namespace == SPRAY:
            spray_data = payload.get('spray')
            self._update_state(space='spray', data=spray_data)

            for el in spray_data:
                evt = HumidifierSpryEvent(device=self,
                                          spry_mode=SprayMode(el.get('mode')),
                                          channel=el.get('channel'),
                                          generated_by_myself=from_myself)
                self.fire_event(evt)
            return True
        elif namespace == LIGHT:
            data = payload.get('light')
            self._update_state(space='light', data=data)
            evt = HumidifierLightEvent(dev=self,
                                       channel=data.get('channel'),
                                       onoff=data.get('onoff'),
                                       rgb=int_to_rgb(data.get('rgb')),
                                       luminance=data.get('luminance'),
                                       generated_by_myself=from_myself)
            self.fire_event(evt)
            return True
        else:
            l.warn("Unknown event: %s" % namespace)
            return False

    def get_status(self):
        with self._state_lock:
            if self._raw_state is None or \
                    self._raw_state.get('spray') is None or \
                    self._raw_state.get('light') is None:
                self._get_status_impl()
        return self._raw_state

    def set_spray_mode(self, spray_mode, channel=0, callback=None):
        payload = {'spray': {'channel': channel, 'mode': spray_mode.value}}
        self.execute_command('SET', SPRAY, payload, callback=callback)

    def get_spray_mode(self, channel=0):
        spray_elements = self.get_status().get('spray')
        for el in spray_elements:
            if el.get('channel') == channel:
                return SprayMode(el.get('mode'))
        return None

    def get_light_state(self):
        light_state = self.get_status().get('light')
        return light_state

    def turn_on_light(self):
        return self.configure_light(onoff=1)

    def turn_off_light(self):
        return self.configure_light(onoff=0)

    def get_light_color(self):
        light = self.get_status().get('light')
        if light is None:
            return None
        rgb = light.get('rgb')
        return int_to_rgb(rgb)

    def configure_light(self, onoff=None, rgb=None, luminance=100, temperature=-1, gradual=0, channel=0):
        if rgb is not None and temperature != -1:
            l.error("You are trying to set both RGB and luminance values for this bulb. It won't work!")

        # Prepare a basic payload
        payload = {
            'light': {
                'channel': channel,
                'gradual': gradual
            }
        }

        if onoff is not None:
            payload['light']['onoff'] = onoff

        mode = 0
        if self.supports_mode(MODE_RGB) and rgb is not None:
            # Convert the RGB to integer
            color = to_rgb(rgb)
            payload['light']['rgb'] = color
            mode = mode | MODE_RGB

        if self.supports_mode(MODE_LUMINANCE) and luminance != -1:
            payload['light']['luminance'] = luminance
            mode = mode | MODE_LUMINANCE

        if self.supports_mode(MODE_TEMPERATURE) and temperature != -1:
            payload['light']['temperature'] = temperature
            mode = mode | MODE_TEMPERATURE

        payload['light']['capacity'] = mode
        self.execute_command(command='SET', namespace=LIGHT, payload=payload)

    def supports_mode(self, mode):
        return (self.get_abilities().get(LIGHT).get('capacity') & mode) == mode

    def _get_status_impl(self):
        data = self.get_sys_data()['all']
        digest = data.get('digest')
        with self._state_lock:
            self._raw_state = digest
        return digest

    def _update_state(self, space, data):
        with self._state_lock:
            if self._raw_state is None:
                self._raw_state = {}

            if space == 'spray':
                self._raw_state['spray'] = data
            elif space == 'light':
                self._raw_state['light'] = data
            else:
                l.warn("Unsupported namespace %s" % space)




from threading import RLock

from meross_iot.cloud.abilities import *
from meross_iot.cloud.device import AbstractMerossDevice
from meross_iot.cloud.devices.subdevices.generic import GenericSubDevice
from meross_iot.logger import HUB_LOGGER as l


class GenericHub(AbstractMerossDevice):
    # Handles the state of this specific HUB
    _state = {}
    _sub_devices = {}
    _subdev_lock = None

    def __init__(self, cloud_client, device_uuid, **kwords):
        super(GenericHub, self).__init__(cloud_client, device_uuid, **kwords)
        self._subdev_lock = RLock()

    def register_sub_device(self,
                            subdev  # type: GenericSubDevice
                            ):
        with self._subdev_lock:
            if subdev.uuid != self.uuid:
                raise Exception("You cannot register this device to this hub, since it has been assigned a different "
                                "hub device uuid.")
            self._sub_devices[subdev.subdevice_id] = subdev

    def _get_status_impl(self):
        res = {}
        data = self.get_sys_data()['all']
        hub_data = data.get('digest').get('hub')
        res['hub_id'] = hub_data.get('hubId')
        res['mode'] = hub_data.get('mode')
        return res

    def get_status(self):
        with self._state_lock:
            if self._state == {}:
                self._state = self._get_status_impl()
            return self._state

    def _dispatch_event_to_subdevice(self, namespace, data, from_myself):
        target = None
        with self._subdev_lock:
            subdevice_id = data.get('id')
            target = self._sub_devices.get(subdevice_id)
            if target is None:
                return

            # Remove the id from the data payload as it will be stored as raw state from the device handler
            del data['id']
            target._handle_push_notification(namespace=namespace, payload=data, from_myself=from_myself)

    def _handle_push_notification(self, namespace, payload, from_myself=False):
        with self._state_lock:
            if namespace == HUB_ONLINE:
                for sensor_data in payload['online']:
                    self._dispatch_event_to_subdevice(namespace=namespace, data=sensor_data, from_myself=from_myself)
                return True

            if namespace == HUB_TOGGLEX:
                for sensor_data in payload['togglex']:
                    self._dispatch_event_to_subdevice(namespace=namespace, data=sensor_data, from_myself=from_myself)
                return True

            elif namespace == HUB_EXCEPTION:
                for ex in payload['exception']:
                    self._dispatch_event_to_subdevice(namespace=namespace, data=ex, from_myself=from_myself)
                return True

            elif namespace == REPORT:
                l.info("Report event is currently unhandled")
                return False

            elif namespace == HUB_MTS100_MODE:
                for sensor_data in payload['mode']:
                    self._dispatch_event_to_subdevice(namespace=namespace, data=sensor_data, from_myself=from_myself)
                return True

            elif namespace == HUB_MTS100_TEMPERATURE:
                for sensor_data in payload['temperature']:
                    self._dispatch_event_to_subdevice(namespace=namespace, data=sensor_data, from_myself=from_myself)
                return True

            else:
                l.error("Unknown/Unsupported namespace/command: %s" % namespace)
                l.debug("Namespace: %s, Data: %s" % (namespace, payload))
                return False

    def _togglex(self, subdevice_id, status, channel=0, callback=None):
        payload = {'togglex': [{'id': subdevice_id, "onoff": status, "channel": channel}]}
        return self.execute_command('SET', HUB_TOGGLEX, payload, callback=callback)

    def get_subdevice_state(self, subdevice_id):
        return self._sub_devices.get(subdevice_id)

    def get_subdevices(self):
        return self._sub_devices.keys()

    def get_subdevice(self, subdevice_id):
        return self._sub_devices.get(subdevice_id)

    def turn_on_subdevice(self, subdevice_id, channel=0, callback=None):
        return self._togglex(subdevice_id, 1, channel, callback=callback)

    def turn_off_subdevice(self, subdevice_id, channel=0, callback=None):
        return self._togglex(subdevice_id, 0, channel, callback=callback)

    def __str__(self):
        self.get_status()
        base_str = super().__str__()
        with self._state_lock:
            if not self.online:
                return base_str
            # TODO: fix this method. We'd probably want to print some more meaningful info
            return "%s [ %s ]" % (base_str, ",".join(self.get_subdevices()))



from threading import RLock

from meross_iot.api import MerossHttpClient
from meross_iot.cloud.abilities import BIND, UNBIND, REPORT, ONLINE
from meross_iot.cloud.client import MerossCloudClient
from meross_iot.cloud.client_status import ClientStatus
from meross_iot.cloud.device_factory import build_wrapper, build_subdevice_wrapper
from meross_iot.cloud.devices.hubs import GenericHub
from meross_iot.logger import MANAGER_LOGGER as l
from meross_iot.meross_event import DeviceOnlineStatusEvent


class MerossManager(object):
    # HTTPClient object used to discover devices
    _http_client = None

    # Dictionary of devices that are currently handled by this manager
    # as UUID -> Device
    _devices = None

    # Lock item used to protect access to the device collection
    _devices_lock = None

    # Cloud credentials to be used against the Meross MQTT cloud
    _cloud_creds = None

    _cloud_client = None

    # List of callbacks that should be called when an event occurs
    _event_callbacks = None
    _event_callbacks_lock = None

    def __init__(self, meross_email, meross_password):
        self._devices_lock = RLock()
        self._devices = dict()
        self._event_callbacks_lock = RLock()
        self._event_callbacks = []

        self._http_client = MerossHttpClient(email=meross_email, password=meross_password)
        self._cloud_creds = self._http_client.get_cloud_credentials()

        # Instantiate the mqtt cloud client
        self._cloud_client = MerossCloudClient(cloud_credentials=self._cloud_creds,
                                               push_message_callback=self._dispatch_push_notification)
        self._cloud_client.connection_status.register_connection_event_callback(callback=self._fire_event)

    def start(self):
        # Connect to the mqtt broker
        self._cloud_client.connect()
        self._discover_devices()

    def stop(self):
        self._cloud_client.close()

    def register_event_handler(self, callback):
        with self._event_callbacks_lock:
            if callback in self._event_callbacks:
                pass
            else:
                self._event_callbacks.append(callback)

    def unregister_event_handler(self, callback):
        with self._event_callbacks_lock:
            if callback not in self._event_callbacks:
                pass
            else:
                self._event_callbacks.remove(callback)

    def get_device_by_uuid(self, uuid):
        self._ensure_started()

        dev = None
        with self._devices_lock:
            dev = self._devices.get(uuid)

        return dev

    def get_device_by_name(self, name):
        self._ensure_started()

        with self._devices_lock:
            for k, v in self._devices.items():
                if v.name.lower() == name.lower():
                    return v
        return None

    def get_supported_devices(self):
        self._ensure_started()
        return [x for k, x in self._devices.items()]

    def get_devices_by_kind(self, clazz):
        self._ensure_started()
        res = []
        with self._devices_lock:
            for k, v in self._devices.items():
                if isinstance(v, clazz):
                    res.append(v)
        return res

    def get_devices_by_type(self, type_name):
        self._ensure_started()
        res = []
        with self._devices_lock:
            for k, v in self._devices.items():
                if v.type.lower() == type_name.lower():
                    res.append(v)
        return res

    def _dispatch_push_notification(self, message, from_myself=False):
        """
        When a push notification is received from the MQTT client, it needs to be delivered to the
        corresponding device. This method serves that scope.
        :param message:
        :param from_myself: boolean flag. When True, it means that the message received is related to a
        previous request issued by this client. When is false, it means the message is related to some other
        client.
        :return:
        """
        header = message['header']      # type: dict
        payload = message['payload']    # type: dict

        # Identify the UUID of the target device by looking at the FROM field of the message header
        dev_uuid = header['from'].split('/')[2]

        # Let's intercept the Bind events: they are useful to trigger new device discovery
        namespace = header.get('namespace')
        if namespace is not None and namespace in [BIND, ONLINE]:
            self._discover_devices()

        # Let's find the target of the event so that it can handle accordingly
        device = None
        with self._devices_lock:
            device = self._devices.get(dev_uuid)

        if device is not None:
            device.handle_push_notification(namespace, payload, from_myself=from_myself)
        else:
            # If we receive a push notification from a device that is not yet contained into our registry,
            # it probably means a new one has just been registered with the meross cloud.
            # This should never happen as we handle the "bind" event just above. However, It appears that the report
            # event is fired before the BIND event. So we need to ignore it at this point
            if namespace == REPORT:
                l.info("Ignoring REPORT event from device %s" % dev_uuid)
            else:
                l.warn("An event was received from an unknown device, which did not trigger the BIND event. "
                       "Device UUID: %s, namespace: %s." % (dev_uuid, namespace))

    def _discover_devices(self, online_only=False):
        """
        Discovers the devices that are visible via HTTP API and update the internal list of
        managed devices accordingly.
        :return:
        """
        for dev in self._http_client.list_devices():
            online = dev['onlineStatus']

            if online_only and online != 1:
                # The device is not online, so we skip it.
                continue

            # If the device we have discovered is not in the list we already handle, we need to add it.
            discovered = self._handle_device_discovered(dev)

            # If the specific device is an HUB, add all its sub devices
            if isinstance(discovered, GenericHub):
                for subdev in self._http_client.list_hub_subdevices(discovered.uuid):
                    self._handle_device_discovered(dev=subdev, parent_hub=discovered)

        return self._devices

    def _handle_device_unbound(self, dev_uuid):
        with self._devices_lock:
            if dev_uuid in self._devices:
                l.info("Unregistering device %s" % dev_uuid)
                del self._devices[dev_uuid]

    def _handle_device_discovered(self, dev, parent_hub=None):
        device = None

        # Check whether we are dealing with a full device or with a subdevice
        if 'deviceType' in dev and 'uuid' in dev:
            # FULL DEVICE case
            d_type = dev['deviceType']
            d_id = dev['uuid']
            device_id = d_id
            device = build_wrapper(device_type=d_type, device_uuid=d_id,
                                   cloud_client=self._cloud_client, device_specs=dev)
        elif 'subDeviceType' in dev and 'subDeviceId' in dev:
            # SUB DEVICE case
            d_type = dev['subDeviceType']
            d_id = dev['subDeviceId']
            device_id = "%s:%s" % (parent_hub.uuid, d_id)
            device = build_subdevice_wrapper(device_type=d_type, device_id=d_id, parent_hub=parent_hub,
                                   cloud_client=self._cloud_client, device_specs=dev)
        else:
            l.warn("Discovered device does not seem to be either a full device nor a subdevice.")
            return

        if device is not None:
            # Check if the discovered device is already in the list of handled devices.
            # If not, add it right away. Otherwise, ignore it.
            is_new = False
            with self._devices_lock:
                if d_id not in self._devices:
                    is_new = True
                    self._devices[device_id] = device

            # If this is new device, register the event handler for it and fire the ONLINE event.
            if is_new:
                with self._event_callbacks_lock:
                    for c in self._event_callbacks:
                        device.register_event_callback(c)

                evt = DeviceOnlineStatusEvent(device, device.online)
                self._fire_event(evt)

        return device

    def _fire_event(self, eventobj):
        for c in self._event_callbacks:
            try:
                c(eventobj)
            except:
                l.exception("An unhandled error occurred while invoking callback")

    def _ensure_started(self):
        if not self._cloud_client.connection_status.check_status(ClientStatus.SUBSCRIBED):
            l.warn("The manager is not connected to the mqtt broker. Did you start the Meross manager?")



from meross_iot.cloud.devices.door_openers import GenericGarageDoorOpener
from meross_iot.cloud.devices.hubs import GenericHub
from meross_iot.cloud.devices.light_bulbs import GenericBulb
from meross_iot.cloud.devices.power_plugs import GenericPlug
from meross_iot.cloud.devices.subdevices.generic import GenericSubDevice
from meross_iot.cloud.devices.subdevices.thermostats import ValveSubDevice
from meross_iot.cloud.devices.humidifier import GenericHumidifier


def build_wrapper(
        cloud_client,
        device_type,  # type: str
        device_uuid,  # type: str
        device_specs  # type: dict
):
    if device_type.startswith('msl') or device_type.startswith('mss560m') or device_type.startswith('mss570m'):
        return GenericBulb(cloud_client, device_uuid=device_uuid, **device_specs)
    elif device_type.startswith('mss'):
        return GenericPlug(cloud_client, device_uuid=device_uuid, **device_specs)
    elif device_type.startswith('msg'):
        return GenericGarageDoorOpener(cloud_client, device_uuid=device_uuid, **device_specs)
    elif device_type.startswith('msh'):
        return GenericHub(cloud_client, device_uuid=device_uuid, **device_specs)
    elif device_type.startswith('msxh'):
        return GenericHumidifier(cloud_client, device_uuid=device_uuid, **device_specs)
    else:
        return GenericPlug(cloud_client, device_uuid=device_uuid, **device_specs)


def build_subdevice_wrapper(cloud_client,
                            device_type,  # type: str
                            device_id,  # type: str
                            device_specs,  # type: dict
                            parent_hub  # type: GenericHub
                            ):
    if device_type.startswith('mts'):
        return ValveSubDevice(cloud_client, subdevice_id=device_id, parent_hub=parent_hub, **device_specs)
    else:
        return GenericSubDevice(cloud_client, subdevice_id=device_id, parent_hub=parent_hub, **device_specs)



from enum import Enum


class MerossEventType(Enum):
    # Fired when the MQTT client connects/disconnects to the MQTT broker
    CLIENT_CONNECTION = 10
    DEVICE_ONLINE_STATUS = 100
    DEVICE_BIND = 200
    DEVICE_UNBIND = 201
    DEVICE_SWITCH_STATUS = 1000
    DEVICE_BULB_SWITCH_STATE = 2000
    DEVICE_BULB_STATE = 2001
    GARAGE_DOOR_STATUS = 3000
    THERMOSTAT_TEMPERATURE_CHANGE = 5000
    THERMOSTAT_MODE_CHANGE = 5001
    HUMIDIFIER_SPRY_EVENT = 6000
    HUMIDIFIER_LIGHT_EVENT = 6001


class MerossEvent(object):
    event_type = None  # type: MerossEventType

    def __init__(self, event_type):
        self.event_type = event_type


class DeviceBindEvent(MerossEvent):
    def __init__(self, device, bind_data):
        super(DeviceBindEvent, self).__init__(MerossEventType.DEVICE_BIND)
        self.device = device
        self.bind_data = bind_data


class DeviceUnbindEvent(MerossEvent):
    def __init__(self, device):
        super(DeviceUnbindEvent, self).__init__(MerossEventType.DEVICE_UNBIND)
        self.device = device


class ClientConnectionEvent(MerossEvent):
    status = None

    def __init__(self, current_status):
        super(ClientConnectionEvent, self).__init__(MerossEventType.CLIENT_CONNECTION)
        self.status = current_status


class DeviceOnlineStatusEvent(MerossEvent):
    # Pointer to the device object
    device = None

    # Current status of the device
    status = None

    def __init__(self, dev, current_status):
        super(DeviceOnlineStatusEvent, self).__init__(MerossEventType.DEVICE_ONLINE_STATUS)
        self.device = dev
        self.status = "online" if current_status else "offline"


class DeviceSwitchStatusEvent(MerossEvent):
    # Pointer to the device object
    device = None

    # Channel ID where the event occurred
    channel_id = None

    # Current state of the switch where the event occurred
    switch_state = None

    # Indicates id the event was generated by a command issued by the library itself.
    # This is particularly useful in the case the user handler wants only to react
    # to events generated by third parties.
    generated_by_myself = None

    def __init__(self, dev, channel_id, switch_state, generated_by_myself):
        super(DeviceSwitchStatusEvent, self).__init__(MerossEventType.DEVICE_SWITCH_STATUS)
        self.device = dev
        self.channel_id = channel_id
        self.switch_state = switch_state
        self.generated_by_myself = generated_by_myself


class DeviceDoorStatusEvent(MerossEvent):
    # Pointer to the device object
    device = None

    # Current state of the door
    door_state = None

    # Channel related to the door controller
    channel = None

    # Indicates id the event was generated by a command issued by the library itself.
    # This is particularly useful in the case the user handler wants only to react
    # to events generated by third parties.
    generated_by_myself = None

    def __init__(self, dev, channel_id, door_state, generated_by_myself):
        super(DeviceDoorStatusEvent, self).__init__(MerossEventType.GARAGE_DOOR_STATUS)
        self.device = dev
        self.channel = channel_id
        self.door_state = "open" if door_state else "closed"
        self.generated_by_myself = generated_by_myself


class BulbSwitchStateChangeEvent(MerossEvent):
    def __init__(self, dev, channel_id, is_on, generated_by_myself):
        super(BulbSwitchStateChangeEvent, self).__init__(MerossEventType.DEVICE_BULB_SWITCH_STATE)
        self.device = dev
        self.channel = channel_id
        self.is_on = is_on
        self.generated_by_myself = generated_by_myself


class BulbLightStateChangeEvent(MerossEvent):
    def __init__(self, dev, channel_id, light_state, generated_by_myself):
        super(BulbLightStateChangeEvent, self).__init__(MerossEventType.DEVICE_BULB_STATE)
        self.device = dev
        self.channel = channel_id
        self.light_state = light_state
        self.generated_by_myself = generated_by_myself


class ThermostatTemperatureChange(MerossEvent):
    def __init__(self, device, temperature_state, generated_by_myself):
        super(ThermostatTemperatureChange, self).__init__(MerossEventType.THERMOSTAT_TEMPERATURE_CHANGE)
        self.device = device
        self.temperature = temperature_state
        self.generated_by_myself = generated_by_myself


class ThermostatModeChange(MerossEvent):
    def __init__(self, device, mode, generated_by_myself):
        super(ThermostatModeChange, self).__init__(MerossEventType.THERMOSTAT_MODE_CHANGE)
        self.device = device
        self.mode = mode
        self.generated_by_myself = generated_by_myself


class HumidifierSpryEvent(MerossEvent):
    def __init__(self, device, spry_mode, channel, generated_by_myself):
        super(HumidifierSpryEvent, self).__init__(MerossEventType.HUMIDIFIER_SPRY_EVENT)
        self.device = device
        self.spry_mode = spry_mode
        self.channel = channel
        self.generated_by_myself = generated_by_myself


class HumidifierLightEvent(MerossEvent):
    def __init__(self, dev, channel, onoff, rgb, luminance, generated_by_myself):
        super(HumidifierLightEvent, self).__init__(MerossEventType.HUMIDIFIER_LIGHT_EVENT)
        self.device = dev
        self.channel = channel
        self.is_on = onoff == 1
        self.rgb = rgb
        self.luminance = luminance
        self.generated_by_myself = generated_by_myself


from meross_iot.cloud.abilities import HUB_BATTERY
from meross_iot.cloud.device import AbstractMerossDevice, HUB_MTS100_ALL, HUB_ONLINE
from meross_iot.meross_event import DeviceOnlineStatusEvent


class GenericSubDevice(AbstractMerossDevice):
    def __init__(self, cloud_client, subdevice_id, parent_hub, **kwords):
        super().__init__(cloud_client, parent_hub.uuid, **kwords)
        self.subdevice_id = subdevice_id
        self._hub = parent_hub
        self.name = kwords.get('subDeviceName')
        self.type = kwords.get('subDeviceType')
        self._hub.register_sub_device(self)
        self._raw_state = {}

    def _get_property(self, parent, child, trigger_update_if_unavailable=True):
        prop = self._raw_state.get(parent, {}).get(child)
        if prop is None and trigger_update_if_unavailable and self.online:
            self._sync_status()
            prop = self._raw_state.get(parent, {}).get(child)
        return prop

    def get_battery_status(self):
        payload = {'battery': [{'id': self.subdevice_id}]}
        data = self.execute_command('GET', HUB_BATTERY, payload).get('battery')[0]
        return data.get('value')

    @property
    def last_active_time(self):
        return self._get_property('online', 'lastActiveTime', trigger_update_if_unavailable=False)

    @property
    def online(self):
        # In any case, if the hub is offline, for sure the device is offline.
        if not self._hub.online:
            return False

        # Subdevices do not report the online status to the HTTP API and they only talk with hubs.
        # So, the idea is to rely to the last_active_time - if available, assuming any interaction < 1 minute ago
        # means the device is still online. If the last_active_time is older, then we try to "ping" the subdevice.
        last_active_on = self.last_active_time
        if last_active_on is None:
            self._sync_status()
        online = self._raw_state.get('online', {}).get('status', 0)
        return online == 1

    def _sync_status(self):
        payload = {'all': [{'id': self.subdevice_id}]}
        res = self._hub.execute_command('GET', HUB_MTS100_ALL, payload)
        data = res.get('all')
        for device_data in data:
            if device_data.get('id') == self.subdevice_id:
                self._raw_state.update(device_data)
        return self._raw_state

    def get_status(self):
        if self._raw_state == {}:
            return self._sync_status()
        else:
            return self._raw_state

    def _handle_push_notification(self, namespace, payload, from_myself=False):
        # The only "common" push notification for all sub-devices seems to be the HUB_ONLINE.
        # We handle this one only for generic sub devices.
        if namespace == HUB_ONLINE:
            online = self._raw_state.get('online')
            if online is None:
                online = {}
                self._raw_state['online'] = online
            online.update(payload)
            online_status = online.get('status')==1
            evt = DeviceOnlineStatusEvent(dev=self, current_status=online_status)
            self.fire_event(evt)
            return True

    def __str__(self):
        return "{}".format(self._raw_state)



class InvalidSignatureException(Exception):
    def __init__(self, message, expected_signature, provided_signature, data):
        super(InvalidSignatureException, self).__init__(message)
        self.expected_signature = expected_signature
        self.provided_signature = provided_signature
        self.data = data
